import datetime
from collections.abc import Mapping
from typing import Any, NamedTuple

from immutables import Map
from typing_extensions import Self, override

from HABApp.core.const import MISSING
from HABApp.core.items import BaseValueItem
from HABApp.core.lib.funcs import compare as _compare
from HABApp.openhab.connection.plugins import send_websocket_event
from HABApp.openhab.interface_sync import get_persistence_data
from HABApp.openhab.items._event_builder import OutgoingCommandEvent, OutgoingStateEvent


class MetaData(NamedTuple):
    value: str
    config: Mapping[str, Any] = Map()


class OpenhabItem(BaseValueItem):
    """Base class for items which exists in OpenHAB.

    :ivar str name:
    :ivar Any value:
    :ivar str | None label:
    :ivar frozenset[str] tags:
    :ivar frozenset[str] groups:
    :ivar Mapping[str, MetaData] metadata:
    """

    _update_to_oh: OutgoingStateEvent
    _command_to_oh: OutgoingCommandEvent

    def __init__(self, name: str, initial_value: Any = None,
                 label: str | None = None, tags: frozenset[str] = frozenset(), groups: frozenset[str] = frozenset(),
                 metadata: Mapping[str, MetaData] = Map()) -> None:
        super().__init__(name, initial_value)
        self.label: str | None = label
        self.tags: frozenset[str] = tags
        self.groups: frozenset[str] = groups
        self.metadata: Mapping[str, MetaData] = metadata

    def _update_item_definition(self, item: 'OpenhabItem') -> None:
        self.label = item.label
        self.tags = item.tags
        self.groups = item.groups
        self.metadata = item.metadata

    @classmethod
    def from_oh(cls, name: str, value: Any = None,
                label: str | None = None, tags: frozenset[str] = frozenset(), groups: frozenset[str] = frozenset(),
                metadata: Mapping[str, MetaData] = Map(), **kwargs: Any) -> Self:
        if value is not None:
            value = cls._state_from_oh_str(value)
        return cls(name, value, label=label, tags=tags, groups=groups, metadata=metadata, **kwargs)

    @staticmethod
    def _state_from_oh_str(state: str):
        """Gets called to convert the state if it is not None"""
        raise NotImplementedError()

    def oh_send_command(self, value: Any = MISSING) -> None:
        """Send a command to the openHAB item

        :param value: (optional) value to be sent. If not specified the current item value will be used.
        """
        new_value = self.value if value is MISSING else value
        send_websocket_event(self._command_to_oh.create_event(self._name, new_value))

    # For the openhab items HABApp internal commands make not much sense
    # so we send the commands to openHAB
    @override
    def command_value(self, value: Any) -> None:
        """Send a command to the openHAB item, the same as oh_send_command

        :param value: value to be sent
        """
        self.oh_send_command(value)

    def oh_post_update(self, value: Any = MISSING) -> None:
        """Post an update to the openHAB item

        :param value: (optional) value to be posted. If not specified the current item value will be used.
        """
        new_value = self.value if value is MISSING else value
        send_websocket_event(self._update_to_oh.create_event(self._name, new_value))

    def oh_post_update_if(self, new_value, *, equal=MISSING, eq=MISSING, not_equal=MISSING, ne=MISSING,
                          lower_than=MISSING, lt=MISSING, lower_equal=MISSING, le=MISSING,
                          greater_than=MISSING, gt=MISSING, greater_equal=MISSING, ge=MISSING,
                          is_=MISSING, is_not=MISSING) -> bool:
        """
        Post a value depending on the current state of the item. If one of the comparisons is true the new state
        will be posted.

        :param new_value: new value to post
        :param equal: item state has to be equal to the passed value
        :param eq: item state has to be equal to the passed value
        :param not_equal: item state has to be not equal to the passed value
        :param ne: item state has to be not equal to the passed value
        :param lower_than: item state has to be lower than the passed value
        :param lt: item state has to be lower than the passed value
        :param lower_equal: item state has to be lower equal the passed value
        :param le: item state has to be lower equal the passed value
        :param greater_than: item state has to be greater than the passed value
        :param gt: item state has to be greater than the passed value
        :param greater_equal: item state has to be greater equal the passed value
        :param ge: item state has to be greater equal the passed value
        :param is_: item state has to be the same object as the passt value (e.g. None)
        :param is_not: item state has to be not the same object as the passt value (e.g. None)

        :return: `True` if the new value was posted else `False`
        """

        if _compare(self.value, equal=equal, eq=eq, not_equal=not_equal, ne=ne,
                    lower_than=lower_than, lt=lt, lower_equal=lower_equal, le=le,
                    greater_than=greater_than, gt=gt, greater_equal=greater_equal, ge=ge, is_=is_, is_not=is_not):
            self.oh_post_update(new_value)
            return True
        return False

    def get_persistence_data(self, persistence: str | None = None,
                             start_time: datetime.datetime | None = None,
                             end_time: datetime.datetime | None = None):
        """Query historical data from the OpenHAB persistence service

        :param persistence: name of the persistence service (e.g. ``rrd4j``, ``mapdb``). If not set default will be used
        :param start_time: return only items which are newer than this
        :param end_time: return only items which are older than this
        """

        return get_persistence_data(
            self._name, persistence, start_time, end_time
        )


HINT_TYPE_OPENHAB_ITEM = type[OpenhabItem]



from typing import Any

from pydantic import BaseModel, Field, TypeAdapter

from HABApp.openhab.definitions import ThingStatusDetailEnum, ThingStatusEnum


class ChannelResp(BaseModel):
    # ChannelDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/dto/ChannelDTO.java

    uid: str
    id: str
    channel_type: str | None = Field(default=None, alias='channelTypeUID')
    item_type: str | None = Field(default=None, alias='itemType')
    kind: str
    label: str = ''
    description: str = ''
    default_tags: tuple[str, ...] = Field(default=tuple, alias='defaultTags')
    properties: dict[str, Any] = {}
    configuration: dict[str, Any] = {}
    auto_update_policy: str = Field(default='', alias='autoUpdatePolicy')

    # EnrichedChannelDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/thing/EnrichedChannelDTO.java
    linked_items: tuple[str, ...] = Field(alias='linkedItems')


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/ThingStatusInfo.java
class ThingStatusResp(BaseModel):
    status: ThingStatusEnum
    detail: ThingStatusDetailEnum = Field(alias='statusDetail')
    description: str | None = None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/firmware/dto/FirmwareStatusDTO.java
class FirmwareStatusResp(BaseModel):
    status: str
    updatable_version: str | None = Field(default=None, alias='updatableVersion')


class ThingResp(BaseModel):
    # AbstractThingDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/dto/AbstractThingDTO.java
    label: str = ''
    bridge_uid: str | None = Field(default=None, alias='bridgeUID')
    configuration: dict[str, Any] = {}
    properties: dict[str, str] = {}
    uid: str = Field(alias='UID')
    thing_type: str = Field(alias='thingTypeUID')
    location: str = ''

    # EnrichedThingDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/thing/EnrichedThingDTO.java
    channels: tuple[ChannelResp, ...] = []
    status: ThingStatusResp = Field(alias='statusInfo')
    firmware_status: FirmwareStatusResp | None = None
    editable: bool


ThingRespList = TypeAdapter(tuple[ThingResp, ...])



from __future__ import annotations

from typing import Generic, TypeVar

from HABApp.core.const.const import MISSING, _MissingType


T = TypeVar('T')


class ValueChange(Generic[T]):
    __slots__ = ('_value', 'changed')

    def __init__(self) -> None:
        self._value: T | _MissingType = MISSING
        self.changed: bool = False

    def set_value(self, value: T):
        current = self._value

        if value is MISSING and current is MISSING:
            self.changed = False
            return self

        if (value is MISSING and current is not MISSING) or (value is not MISSING and current is MISSING):
            self._value = value
            self.changed = True
            return self

        if value != current:
            self._value = value
            self.changed = True
            return self

        self.changed = False
        return self

    def set_missing(self):
        self.set_value(MISSING)
        return self

    @property
    def is_missing(self) -> bool:
        return self._value is MISSING

    @property
    def value(self) -> T:
        if self._value is MISSING:
            raise ValueError()
        return self._value

    def __repr__(self) -> str:
        now = self._value if self._value is not MISSING else repr(MISSING)
        return f'<{self.__class__.__name__} value: {now} changed: {self.changed}>'



import logging
from collections.abc import Callable, Coroutine
from typing import Any, Final

from typing_extensions import override

from HABApp.core.asyncio import create_task
from HABApp.core.internals import Context
from HABApp.core.internals.wrapped_function.base import P, R, WrappedFunctionBase


class WrappedAsyncFunction(WrappedFunctionBase[P, R]):

    def __init__(self, coro: Callable[P, Coroutine[Any, Any, R]],
                 name: str | None = None,
                 logger: logging.Logger | None = None,
                 context: Context | None = None) -> None:

        super().__init__(name=name, func=coro, logger=logger, context=context)

        self.coro: Final = coro

    @override
    def run(self, *args: P.args, **kwargs: P.kwargs) -> None:
        create_task(self.async_run(*args, **kwargs), name=self.name)

    @override
    async def async_run(self, *args: P.args, **kwargs: P.kwargs) -> R | None:

        try:
            return await self.coro(*args, **kwargs)
        except Exception as e:
            self.process_exception(e, *args, **kwargs)
            return None



from __future__ import annotations

import asyncio
import logging
from asyncio import sleep
from pathlib import Path
from time import monotonic
from typing import TYPE_CHECKING, Final

from pydantic import ValidationError

import HABApp
from HABApp.core.const.topics import TOPIC_FILES
from HABApp.core.files.file import HABAppFile
from HABApp.core.files.file_properties import get_file_properties
from HABApp.core.files.name_builder import FileNameBuilder
from HABApp.core.lib import SingleTask


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from re import Pattern

    from HABApp.core.events.habapp_events import RequestFileLoadEvent, RequestFileUnloadEvent
    from HABApp.core.files.watcher import HABAppFileWatcher


log = logging.getLogger('HABApp.files')


class FileTypeHandler:
    def __init__(self, name: str, logger: logging.Logger, *,
                 prefix: str, pattern: Pattern | None = None,
                 on_load: Callable[[str, Path], Awaitable[None]],
                 on_unload: Callable[[str, Path], Awaitable[None]]) -> None:
        self.name: Final = name
        self.logger: Final = logger

        self.prefix: Final = prefix
        self.pattern: Final = pattern

        self.on_load: Final = on_load
        self.on_unload: Final = on_unload

    def matches(self, name: str) -> bool:
        if not name.startswith(self.prefix):
            return False

        if (p := self.pattern) is not None and not p.search(name):  # noqa: SIM103
            return False

        return True

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.name:s}>'


class FileManager:
    def __init__(self, watcher: HABAppFileWatcher | None) -> None:
        self._lock = asyncio.Lock()
        self._files: Final[dict[str, HABAppFile]] = {}

        self._files_request_load: Final[set[str]] = set()
        self._files_request_unload: Final[set[str]] = set()

        self._file_names: Final = FileNameBuilder()
        self._file_handlers: tuple[FileTypeHandler, ...] = ()

        self._task: Final = SingleTask(self._load_file_task, name='file load worker')
        self._watcher: Final = watcher

        self._event_received: bool = False

    def add_folder(self, prefix: str, folder: Path, *,
                   name: str, priority: int, pattern: Pattern | None = None) -> None:
        self._file_names.add_folder(prefix, folder, priority=priority, pattern=pattern)
        if self._watcher is not None:
            self._watcher.watch_folder(name, self.file_watcher_event, folder)

    def get_file_watcher(self) -> HABAppFileWatcher:
        if self._watcher is None:
            raise ValueError()
        return self._watcher

    def get_folders(self):  # noqa: ANN201
        return self._file_names.get_folders()

    def add_handler(self, name: str, logger: logging.Logger, *,
                   prefix: str, pattern: Pattern | None = None,
                   on_load: Callable[[str, Path], Awaitable[None]],
                   on_unload: Callable[[str, Path], Awaitable[None]]) -> None:

        for h in self._file_handlers:
            if h.name == name:
                msg = f'Handler {name:s} already exists!'
                raise ValueError(msg)
            if h.prefix == prefix and h.pattern == pattern:
                msg = f'Handler with prefix {prefix:s} and pattern {pattern} already exists!'
                raise ValueError(msg)

        new = FileTypeHandler(name, logger, prefix=prefix, pattern=pattern, on_load=on_load, on_unload=on_unload)
        self._file_handlers += (new, )
        log.debug(f'Added handler {new.name}')

    def get_file(self, name: str) -> HABAppFile | None:
        return self._files.get(name)

    def file_state_changed(self, file: HABAppFile, new_state: str) -> None:
        log.debug(f'{file.name} changed to {new_state:s}')
        for f in self._files.values():
            f.file_state_changed(file, self)

    def _get_file_handler(self, name: str) -> FileTypeHandler:
        handlers = [h for h in self._file_handlers if h.matches(name)]
        if not handlers:
            msg = f'No handler matched for {name:s}'
            raise ValueError(msg)

        if len(handlers) > 1:
            msg = f'Multiple handlers matched for {name:s}: {", ".join(str(h) for h in handlers)}'
            raise ValueError(msg)

        return handlers[0]

    async def _do_file_load(self, name: str) -> None:
        if not (file := self.get_file(name)):
            return None
        await file.load(self._get_file_handler(name), manager=self)

    async def _do_file_unload(self, name: str) -> None:
        if not (file := self.get_file(name)):
            return None
        await file.unload(self._get_file_handler(name), manager=self)

        if file.can_be_removed():
            self._files.pop(name)

    async def _load_file_task(self, *, keep_alive: bool = True) -> None:
        try:
            task_sleep = 0.4
            task_alive = 15

            task_shutdown = False
            last_process = monotonic()

            while True:
                await sleep(0)

                # wait to aggregate changes
                while True:
                    async with self._lock:
                        if not self._event_received:
                            break
                        self._event_received = False

                    await sleep(task_sleep)

                async with self._lock:
                    # we first try to unload all files
                    if self._files_request_unload:
                        # unload order is reverse of load order, since we unload unconditionally we can
                        unload_name = next(self._file_names.get_names(self._files_request_unload, reverse=True))
                        self._files_request_unload.remove(unload_name)
                        await self._do_file_unload(unload_name)
                        last_process = monotonic()
                        continue

                    # then we add all the files we want to load
                    if self._files_request_load:
                        load_name = next(self._file_names.get_names(self._files_request_load))
                        if (existing := self.get_file(load_name)) and existing.can_be_unloaded():
                            self._files_request_unload.add(load_name)
                            continue
                        self._files_request_load.remove(load_name)
                        self._files[load_name] = self.__create_file(load_name)
                        last_process = monotonic()
                        continue

                    # Once we cleared all the queues the proper processing starts:
                    # check files for dependencies etc.
                    for file in self._files.values():
                        file.check_properties(self, log, log_msg=task_shutdown)
                        file.check_dependencies(self)

                    # unload pending files
                    if unload_pending := [f.name for f in self._files.values() if f.state_unload_pending()]:
                        name = next(self._file_names.get_names(unload_pending, reverse=True))
                        await self._do_file_unload(name)
                        last_process = monotonic()
                        continue

                    # load files
                    if can_be_loaded := [f.name for f in self._files.values() if f.can_be_loaded()]:
                        name = next(self._file_names.get_names(can_be_loaded))
                        await self._do_file_load(name)
                        last_process = monotonic()
                        continue

                if task_shutdown or not keep_alive:
                    break
                await sleep(0.1)
                task_shutdown = monotonic() - last_process > task_alive

        except Exception as e:
            HABApp.core.wrapper.process_exception(self._task.name, e, logger=log)
        log.debug('Worker done!')

    def __accept_event(self, event: RequestFileLoadEvent | RequestFileUnloadEvent) -> bool:
        if not self._file_names.is_accepted_name(event.name):
            HABApp.core.logger.log_error(log, f'Ignoring {event.__class__.__name__} for invalid name "{event.name}"')
            return False
        return True

    def __create_file(self, name: str) -> HABAppFile:
        path = self._file_names.create_path(name)
        text = path.read_text()
        checksum = HABAppFile.create_checksum(text)
        try:
            properties = get_file_properties(text)
        except ValidationError as e:
            properties = None
            HABApp.core.logger.log_error(log, str(e))

        habapp_file = HABAppFile(name, path, checksum, properties)
        log.debug(f'{habapp_file.name} created in state {habapp_file._state:s}')
        return habapp_file

    async def event_load(self, event: RequestFileLoadEvent) -> None:
        if not self.__accept_event(event):
            return None

        self._task.start_if_not_running()

        async with self._lock:
            self._event_received = True
            self._files_request_load.add(event.name)

    async def event_unload(self, event: RequestFileUnloadEvent) -> None:
        if not self.__accept_event(event):
            return None

        self._task.start_if_not_running()

        async with self._lock:
            self._event_received = True
            self._files_request_unload.add(event.name)

    async def file_watcher_event(self, path: str) -> None:
        if not self._file_names.is_accepted_path(path):
            return None

        obj = Path(path)
        name = self._file_names.create_name(path)

        if obj.is_dir():
            return None

        if not obj.is_file():
            HABApp.core.EventBus.post_event(TOPIC_FILES, HABApp.core.events.habapp_events.RequestFileUnloadEvent(name))
            return None

        if existing := self.get_file(name):
            checksum = HABAppFile.create_checksum(obj.read_text())
            if existing.checksum == checksum:
                log.debug(f'Skip file system event because file {name:s} did not change')
                return None

        HABApp.core.EventBus.post_event(TOPIC_FILES, HABApp.core.events.habapp_events.RequestFileLoadEvent(name))
        return None

    def setup(self) -> None:
        HABApp.core.EventBus.add_listener(
            HABApp.core.internals.EventBusListener(
                TOPIC_FILES, HABApp.core.internals.wrap_func(self.event_load),
                HABApp.core.events.EventFilter(HABApp.core.events.habapp_events.RequestFileLoadEvent)
            )
        )

        HABApp.core.EventBus.add_listener(
            HABApp.core.internals.EventBusListener(
                TOPIC_FILES, HABApp.core.internals.wrap_func(self.event_unload),
                HABApp.core.events.EventFilter(HABApp.core.events.habapp_events.RequestFileUnloadEvent)
            )
        )



import logging
import random
import string
from collections.abc import Generator
from pathlib import Path
from typing import Literal

import pydantic
from easyconfig.models import BaseModel
from pydantic import Field


QOS = Literal[0, 1, 2]


class TLSSettings(BaseModel):
    enabled: bool = Field(default=True, description='Enable TLS for the connection')
    ca_cert: Path = Field(
        default='', description='Path to a CA certificate that will be treated as trusted', alias='ca cert')
    insecure: bool = Field(
        default=False, description='Validate server hostname in server certificate')


class Connection(BaseModel):
    identifier: str = Field('HABApp-' + ''.join(random.choices(string.ascii_letters, k=13)),
                            description='Identifier that is used to uniquely identify this client on the mqtt broker.')
    host: str = Field('', description='Connect to this host. Empty string ("") disables the connection.')
    port: int = 1883
    user: str = ''
    password: str = ''
    tls: TLSSettings = Field(default_factory=TLSSettings)

    # implemented 2024.02.0
    @pydantic.model_validator(mode='before')
    @classmethod
    def _migrate_client_id(cls, data):
        if isinstance(data, dict) and 'client_id' in data:
            log = logging.getLogger('HABApp.Config')
            log.warning('"client_id" in mqtt.connection has been renamed to "identifier"')
            if 'identifier' not in data:
                data['identifier'] = data.pop('client_id')
        return data


class Subscribe(BaseModel):
    qos: QOS = Field(default=0, description='Default QoS for subscribing')
    topics: tuple[str | tuple[str, QOS], ...] = Field(default=('#', 'topic/with/default/qos', ('topic/with/qos', 1)))

    def get_topic_qos(self) -> Generator[tuple[str, QOS], None, None]:
        for obj in self.topics:
            if isinstance(obj, str):
                yield obj, self.qos
            else:
                yield obj

    # Implemented 2024.11.0
    @pydantic.model_validator(mode='before')
    @classmethod
    def _migrate_topics(cls, data):
        if isinstance(data, dict) and (topics := data.get('topics', [])) is not None:
            for i, topic_obj in enumerate(topics):
                if not isinstance(topic_obj, list):
                    continue
                topic, qos = topic_obj
                if qos is not None:
                    continue

                log = logging.getLogger('HABApp.Config')
                log.warning('Empty QoS is not longer allowed for subscribing to topics.')
                log.warning('Specify QOS or remove empty entry, e.g from')
                log.warning(f'  - - {topic:s}')
                log.warning('    - ')
                log.warning('to')
                log.warning(f'  - {topic:s}')

                topics[i] = topic

        return data


class Publish(BaseModel):
    qos: QOS = Field(default=0, description='Default QoS when publishing values')
    retain: bool = Field(default=False, description='Default retain flag when publishing values')


class General(BaseModel):
    listen_only: bool = Field(False, description='If True HABApp does not publish any value to the broker')


class MqttConfig(BaseModel):
    """MQTT configuration"""

    connection: Connection = Field(default_factory=Connection)
    subscribe: Subscribe = Field(default_factory=Subscribe)
    publish: Publish = Field(default_factory=Publish)
    general: General = Field(default_factory=General)



import logging


class BufferedLogger:
    def __init__(self) -> None:
        self._msgs: list[tuple[int, str]] = []

    def _log(self, lvl: int, msg: str) -> None:
        self._msgs.append((lvl, msg))

    def debug(self, msg: str) -> 'BufferedLogger':
        self._log(logging.DEBUG, msg)
        return self

    def info(self, msg: str) -> 'BufferedLogger':
        self._log(logging.INFO, msg)
        return self

    def warning(self, msg: str) -> 'BufferedLogger':
        self._log(logging.WARNING, msg)
        return self

    def error(self, msg: str) -> 'BufferedLogger':
        self._log(logging.ERROR, msg)
        return self

    def flush(self, logger: logging.Logger) -> None:
        for lvl, msg in self._msgs:
            logger.log(lvl, msg)
        self._msgs.clear()
